package net.bigeon.gclc.system;

/*-
 * #%L
 * GCLC system command
 * %%
 * Copyright (C) 2016 - 2018 Bigeon
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */

import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.bigeon.gclc.exception.CommandRunException;
import net.bigeon.gclc.exception.CommandRunExceptionType;
import net.bigeon.gclc.manager.ConsoleOutput;

/** A class forwarding the content of a stream into a console output.
 * <p>
 * This class is useful for processes that run in a different environment than
 * the application but return an output. Such processes include command
 * processes and socket communications for example.
 * <p>
 * In the current library is is used to retrieve a system command output.
 * 
 * @author Emmanuel Bigeon */
public final class ForwardingRunnable implements Runnable {
    /** Error message for the manager closed. */
    private static final String MANAGER_WAS_CLOSED = "manager was closed";
    /** The class logger */
    private static final Logger LOGGER             = Logger
            .getLogger(ForwardingRunnable.class.getName());

    /** The console output. */
    private final ConsoleOutput out;
    /** The input stream. */
    private final InputStream   is;
    private Exception           error;

    /** Create the runnable.
     * 
     * @param out the output console
     * @param is the input stream */
    public ForwardingRunnable(final ConsoleOutput out, final InputStream is) {
        this.out = out;
        this.is = is;
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run() */
    @Override
    public void run() {
        try {
            readToEnd();
            is.close();
        } catch (final CommandRunException e) {
            LOGGER.log(Level.WARNING, "Manager was closed in the meantime...", e); //$NON-NLS-1$
            setError(e);
        } catch (final IOException e) {
            LOGGER.log(Level.WARNING, "Input stream was closed...", e); //$NON-NLS-1$
            setError(e);
        }
    }

    /** Read the input until its end.
     * 
     * @throws CommandRunException if the manager was closed while writing the
     *         stream */
    private void readToEnd() throws CommandRunException {
        int c;
        try {
            while ((c = is.read()) >= 0) {
                // Print the character in the output
                printCharacter(out, (char) c);
            }
        } catch (final IOException e) {
            LOGGER.log(Level.INFO, "input stream reading failed", e); //$NON-NLS-1$
            setError(e);
        }
    }

    /** Print a character on the console.
     * 
     * @param out the console
     * @param c the character
     * @throws CommandRunException if the console failed to print the character. */
    private static void printCharacter(final ConsoleOutput out, final char c)
            throws CommandRunException {
        try {
            out.print(Character.toString(c));
        } catch (final IOException e) {
            throw new CommandRunException(CommandRunExceptionType.INTERACTION,
                    MANAGER_WAS_CLOSED, e); // $NON-NLS-1$
        }
    }

    /** Tool method to connect an input stream to a console output.
     * 
     * @param stream the stream
     * @param output the output
     * @param name the identification of the thread
     * @return the thread the forwarding runnable runs into. */
    public static Thread connect(final InputStream stream, final ConsoleOutput output,
            final String name) {
        final Thread th = new Thread(new ForwardingRunnable(output, stream), name);
        th.start();
        return th;
    }

    /** Tool method to connect an input stream to a console output.
     * 
     * @param stream the stream
     * @param output the output
     * @return the thread the forwarding runnable runs into. */
    public static Thread connect(final InputStream stream, final ConsoleOutput output) {
        return connect(stream, output, "Forwarding");
    }

    /** Get the error in this runnable execution if any.
     * <p>
     * In the current implementation, this error can be an {@link IOException} if
     * the input stream failed or a {@link CommandRunException} if the output was
     * closed.
     * 
     * @return the error */
    public synchronized Exception getError() {
        return error;
    }

    /** Set the error that caused this runnable to stop executing.
     * 
     * @param error the error to set */
    private synchronized void setError(final Exception error) {
        this.error = error;
    }
}
